unit pdfsearcher;

{$mode objfpc}{$H+}
{$modeswitch advancedrecords}

interface

uses
  Classes, SysUtils, DateUtils, db, sqldb, SQLDBindexDB, pqconnection, pgindexdb, fpIndexer, inifiles, httpdefs, fpjson;

Type
  TArticleData = record
    StartPage,
    EndPage : Integer;
    Author,
    Title,
    Issue,
    PDF : String;
    function Match(aPDF : String; aPage : Integer) : boolean;
  end;
  TArticleDataArray = Array of TArticleData;

  { TSearcher }

  TSearcher = Class(TComponent)
  private
    FAllowCors: Boolean;
    FArticles: TArticleDataArray;
    FDB : TCustomIndexDB;
    FPDFLocation: String;
    FSearch : TFPSearch;
    FDefaultMinRank : Integer;
    FMinRank : Integer;
    FFormattedJSON : Boolean;
    FDefaultMetadata,
    FIncludeMetaData : Boolean;
    FDefaultAvailable : TAvailableMatch;
    FMetadata : TJSONObject;
    FWordsMetadata : TJSONObject;
    FConnection : TPQConnection;
    procedure ConfigSearch(aRequest: TRequest; aResponse: TResponse);
    procedure ConfigWordList(aRequest: TRequest; out aContaining : UTF8string; Out Partial : TAvailableMatch; Out aSimple : Boolean);
    procedure ConnectToDatabase;
    procedure DisConnectFromDatabase;
    function FindArticle(Position: Integer; const aFile: String): TArticleData;
    function GetConfigFileName: string;
    function SearchDataToJSON(aID: Integer; const aRes: TSearchWordData): TJSONObject;
    procedure SendJSON(J: TJSONObject; aResponse: TResponse);
    procedure SetupMetadata;
    procedure LoadArticles;
  Protected
    function InitSearch(aResponse: TResponse): Boolean;
    function SetupDB(aIni: TCustomIniFile): TCustomIndexDB;
    Property DB : TCustomIndexDB Read FDB;
    Property Search : TFPSearch Read FSearch;
    Property MinRank : Integer Read FMinRank;
    Property FormattedJSON : Boolean Read FFormattedJSON;
    Property Articles : TArticleDataArray Read FArticles;
  Public
    Procedure Init;
    procedure DoDBLog(Sender: TSQLConnection; EventType: TDBEventType; const Msg: String);
    Function CheckParams(aRequest : TRequest; aResponse : TResponse) : Boolean;
    Function CheckListParams(aRequest : TRequest; aResponse : TResponse) : Boolean;
    Procedure PDFSearch(aRequest : TRequest; aResponse : TResponse);
    Procedure WordList(aRequest : TRequest; aResponse : TResponse);
    Property AllowCors : Boolean Read FAllowCors Write FAllowCors;
    Property PDFLocation : String Read FPDFLocation;
  end;


implementation

function TArticleData.Match(aPDF : String; aPage : Integer) : boolean;
begin
  Result:=SameText(Pdf,aPdf) and (StartPage<=aPage) and (aPage<=EndPage)
end;


function TSearcher.SetupDB(aIni: TCustomIniFile): TCustomIndexDB;

Const
  SDatabase = 'Database';
  KeyHostName = 'HostName';
  KeyDatabaseName = 'DatabaseName';
  KeyUser = 'User';
  KeyPassword = 'Password';

Var
  QDB : TSQLDBIndexDB;

begin
  Result:=nil;
  QDB := TPGIndexDB.Create(nil);
  QDB.HostName:= aIni.ReadString(SDatabase,KeyHostName,QDB.HostName);
  QDB.DatabasePath := aIni.ReadString(SDatabase,KeyDatabaseName,QDB.DatabasePath);
  QDB.UserName := aIni.ReadString(SDatabase,KeyUser,QDB.UserName);
  QDB.Password := aIni.ReadString(SDatabase,KeyPassword,QDB.Password);
  Result:=QDB;
  FConnection:=TPQConnection.Create(Self);
  FConnection.HostName:= aIni.ReadString(SDatabase,KeyHostName,QDB.HostName);
  FConnection.DatabaseName := aIni.ReadString(SDatabase,KeyDatabaseName,QDB.DatabasePath);
  FConnection.UserName := aIni.ReadString(SDatabase,KeyUser,QDB.UserName);
  FConnection.Password := aIni.ReadString(SDatabase,KeyPassword,QDB.Password);
  FConnection.Transaction:=TSQLTransaction.Create(Self);
end;

function TSearcher.CheckParams(aRequest: TRequest; aResponse: TResponse): Boolean;

Var
  S : String;
  B : Boolean;

begin
  S:=aRequest.QueryFields.Values['q'];
  Result:=S<>'';
  if not Result then
    begin
    aResponse.Code:=400;
    aResponse.CodeText:='Missing q param';
    aResponse.SendResponse;
    end;
  S:=aRequest.QueryFields.Values['r'];
  Result:=(S='') or (StrToIntDef(S,-1)<>-1);
  if not Result then
    begin
    aResponse.Code:=400;
    aResponse.CodeText:='Wrong value for r';
    aResponse.SendResponse;
    end;
  S:=aRequest.QueryFields.Values['c'];
  Result:=(S='') or TryStrToBool(S,B);
  if not Result then
    begin
    aResponse.Code:=400;
    aResponse.CodeText:='Wrong value for c';
    aResponse.SendResponse;
    end;
  S:=aRequest.QueryFields.Values['m'];
  Result:=(S='') or TryStrToBool(S,B);
  if not Result then
    begin
    aResponse.Code:=400;
    aResponse.CodeText:='Wrong value for m';
    aResponse.SendResponse;
    end;
end;

function TSearcher.CheckListParams(aRequest: TRequest; aResponse: TResponse): Boolean;

Var
  m,S : String;
  B : Boolean;

begin
  S:=aRequest.QueryFields.Values['q'];
  M:=aRequest.QueryFields.Values['t'];
  Result:=(M='');
  if not Result then
    case lowercase(M) of
      'all' :
        if S<>'' then
          begin
          aResponse.Code:=400;
          aResponse.CodeText:='Q must be empty';
          aResponse.SendResponse;
          end;
      'contains',
      'exact',
      'startswith' :
        if S='' then
          begin
          aResponse.Code:=400;
          aResponse.CodeText:='Q may not be empty';
          aResponse.SendResponse;
          end;
    else
      aResponse.Code:=400;
      aResponse.CodeText:='Wrong value for t';
      aResponse.SendResponse;
    end;
  S:=aRequest.QueryFields.Values['s'];
  Result:=(S='') or TryStrToBool(S,B);
  if not Result then
    begin
    aResponse.Code:=400;
    aResponse.CodeText:='Wrong value for s';
    aResponse.SendResponse;
    end;
  if not B then
    begin
    S:=aRequest.QueryFields.Values['m'];
    Result:=(S='') or TryStrToBool(S,B);
    if not Result then
      begin
      aResponse.Code:=400;
      aResponse.CodeText:='Wrong value for m';
      aResponse.SendResponse;
      end;
    end;
end;

procedure TSearcher.SetupMetadata;

begin
  FMetadata:=TJSONObject.Create([
    'root', 'data',
    'idField','id',
    'fields',TJSONArray.Create([
      TJSONObject.Create(['name','id','type','int']),
      TJSONObject.Create(['name','rank','type','int']),
      TJSONObject.Create(['name','articlePage','type','int']),
      TJSONObject.Create(['name','articleIssue','type','string','maxlen',10]),
      TJSONObject.Create(['name','articleAuthor','type','string','maxlen',127]),
      TJSONObject.Create(['name','articleTitle','type','string','maxlen',255]),
      TJSONObject.Create(['name','url','type','string','maxlen',100]),
      TJSONObject.Create(['name','context','type','string','maxlen',MaxContextLen]),
      TJSONObject.Create(['name','date','type','date'])
     ])
  ]);
  FWordsMetadata:=TJSONObject.Create([
    'root', 'data',
    'idField','id',
    'fields',TJSONArray.Create([
      TJSONObject.Create(['name','id','type','int']),
      TJSONObject.Create(['name','word','type','string','maxlen',100])
     ])
  ]);
end;

procedure TSearcher.ConnectToDatabase;

begin
  FConnection.Open;
end;

procedure TSearcher.DisConnectFromDatabase;

begin
  FConnection.Close(True);
end;

procedure TSearcher.DoDBLog(Sender: TSQLConnection; EventType: TDBEventType;
  const Msg: String);
begin
  Writeln('[',eventType,']: ',Msg);
end;

procedure TSearcher.LoadArticles;

Var
  Q : TSQLQuery;
  I : integer;
  FStartPage, FEndPage, FAuthor, FTitle, FIssue, FPDF : TField;
  Article : TArticleData;


begin
  Q:=Nil;
  ConnectToDatabase;
  try
    Q:=TSQLQuery.Create(Self);
    Q.DataBase:=FConnection;
    Q.Transaction:=FConnection.Transaction;
    Q.SQL.Text:='select * from articles';
    Q.UsePrimaryKeyAsKey:=False;
    Q.Open;
    Q.Last;
    SetLength(FArticles,Q.RecordCount);
    Q.First;
    FStartPage:=Q.FieldByName('startpage');
    FEndPage:=Q.FieldByName('endpage');
    FAuthor:=Q.FieldByName('author');
    FTitle:=Q.FieldByName('title');
    FIssue:=Q.FieldByName('issue');
    FPDF:=Q.FieldByName('pdffile');
    I:=0;
    While not Q.EOF do
       begin
       Article.StartPage:=FStartPage.AsInteger;
       Article.EndPage:=FEndPage.AsInteger;
       Article.Author:=FAuthor.AsString;
       Article.Title:=FTitle.AsString;
       Article.Issue:=FIssue.AsString;
       Article.PDF:=FPDF.AsString;
       Farticles[I]:=Article;
       Inc(I);
       Q.Next;
       end;
  finally
    Q.Free;
    DisconnectFromDatabase;
  end;
end;

function TSearcher.InitSearch(aResponse: TResponse): Boolean;

begin
  if FDB<>Nil then
    exit(True);
end;

function TSearcher.GetConfigFileName : string;

Const
  BaseName ='pdfsearch.ini';

  Function TestCfg(aDir : string) : String;

  begin
    Result:=aDir+BaseName;
    if not FileExists(Result) then
      Result:='';
  end;

Var
  CFN : String;

begin
  CFN:=TestCfg(GetAppConfigDir(true));
  if (CFN='') then
    CFN:=TestCfg(GetAppConfigDir(False));
  if (CFN='') then
    CFN:=TestCfg('config/');
  if (CFN='') then
    CFN:=TestCfg(ExtractFilePath(ParamStr(0)));
  if (CFN='') then
    CFN:=TestCfg('');
  if (CFN='') then
    Raise Exception.Create('No config file found');
  Result:=CFN;
end;

procedure TSearcher.Init;

Const
  DefaultLocation = '/home/michael/Documents/pdf/blaise/';

Var
  CFN : String;
  aIni: TMemIniFile;
begin
  CFN:=GetConfigFileName;
  aIni:=TMemIniFile.Create(CFN);
  try
    FFormattedJSON:=aIni.ReadBool('search','formatjson',False);
    FDefaultMinRank:=aIni.ReadInteger('search','minrank',1);
    FDefaultMetadata:=aIni.ReadBool('search','metadata',true);
    FAllowCors:=aIni.ReadBool('search','allowcors',true);
    FPDFLocation:=aIni.ReadString('Files','PDF',DefaultLocation);
    FDB:=SetupDB(aIni);
    FSearch:=TFPSearch.Create(Self);
    FSearch.Database:=FDB;
  finally
    aIni.Free;
  end;
  SetupMetadata;
end;

procedure TSearcher.ConfigSearch(aRequest: TRequest; aResponse: TResponse);

Var
  S : string;
  O : TSearchOptions;
  B : Boolean;

begin
  FMinRank:=StrToIntDef(aRequest.QueryFields.Values['r'],0);
  if FMinRank=0 then
    FMinRank:=FDefaultMinRank;
  S:=aRequest.QueryFields.Values['m'];
  if (S='') or Not TryStrToBool(S,FIncludeMetaData)  then
    FIncludeMetaData:=FDefaultMetaData;
  FSearch.SetSearchWord(aRequest.QueryFields.Values['q']);
  O:=[];
  S:=aRequest.QueryFields.Values['c'];
  if (S<>'') and TryStrToBool(S,B) and B then
    Include(O,soContains);
  FSearch.Options:=O;
end;

procedure TSearcher.ConfigWordList(aRequest: TRequest; out aContaining: UTF8string; out Partial: TAvailableMatch; out aSimple: Boolean);

Var
  m,S : String;

begin
  aContaining:=aRequest.QueryFields.Values['q'];
  M:=aRequest.QueryFields.Values['t'];
  case lowercase(M) of
    'all' : Partial:=amAll;
    'contains' : Partial:=amContains;
    'exact' : Partial:=amExact;
    'startswith' : Partial:=amStartsWith;
  else
    Partial:=FDefaultAvailable;
    if (Partial<>amAll) and (aContaining='') then
      Partial:=amAll;
  end;
  S:=aRequest.QueryFields.Values['s'];
  if (S='') then
    aSimple:=False
  else
    aSimple:=StrToBool(S);
  if ASimple then
    FIncludeMetadata:=False
  else
    begin
    FIncludeMetaData:=FDefaultMetaData;
    S:=aRequest.QueryFields.Values['m'];
    if (S<>'') then
      TryStrToBool(S,FIncludeMetaData);
    end
end;

function TSearcher.FindArticle(Position: Integer; const aFile: String
  ): TArticleData;

Var
  A : TArticleData;

begin
  Result:=Default(TArticleData);
  if Length(Articles)=0 then
    LoadArticles;
  for A in Articles do
    if A.Match(aFile,Position) then
      Exit(A);
end;

function TSearcher.SearchDataToJSON(aID: Integer; const aRes: TSearchWordData
  ): TJSONObject;

Var
  Article : TArticleData;

begin
  Article:=FindArticle(aRes.Position,aRes.URL);
  Result:=TJSONObject.Create([
    'id',aID,
    'rank',aRes.Rank,
    'articlePage', aRes.Position,
    'articleTitle', Article.Title,
    'articleAuthor', Article.Author,
    'articleIssue', Article.Issue,
    'url',aRes.URL,
    'context',ares.Context,
    'date',FormatDateTime('yyyy"-"mm"-"dd"T"hh":"nn":"ss',aRes.FileDate)
  ]);
end;

procedure TSearcher.PDFSearch(aRequest: TRequest; aResponse: TResponse);

Var
  I : Integer;
  J : TJSONObject;
  A : TJSONArray;

begin
  aResponse.ContentType:='application/json';
  if AllowCORS then
    AResponse.SetCustomHeader('Access-Control-Allow-Origin','*');
  if not CheckParams(aRequest,aResponse) then
    exit;
  if not InitSearch(aResponse) then
    begin
    aResponse.Code:=500;
    aResponse.CodeText:='Internal error';
    aResponse.SendResponse;
    exit;
    end;
  ConfigSearch(aRequest,aResponse);
  FSearch.Execute;
  A:=nil;
  J:=TJSONObject.Create;
  try
    if FIncludeMetadata then
      J.Add('metaData',FMetadata.Clone);
    A:=TJSONArray.Create;
    For I:=0 to Search.RankedCount-1 do
      begin
      if Search.RankedResults[I].Rank>=MinRank then
        A.Add(SearchDataToJSON(I+1,Search.RankedResults[I]));
      end;
    J.Add('data',A);
    SendJSON(J,aResponse);
  finally
    J.Free;
  end;
end;

procedure TSearcher.SendJSON(J : TJSONObject; aResponse: TResponse);

begin
  if FormattedJSON then
    aResponse.Content:=J.FormatJSON()
  else
    aResponse.Content:=J.AsJSON;
  aResponse.ContentLength:=Length(aResponse.Content);
  aResponse.SendContent;
end;

procedure TSearcher.WordList(aRequest: TRequest; aResponse: TResponse);

Var
  I : Integer;
  J : TJSONObject;
  A : TJSONArray;
  w,aContaining : UTF8String;
  aPartial : TAvailableMatch;
  aSimple : Boolean;
  aList : TUTF8StringArray;


begin
  aResponse.ContentType:='application/json';
  if AllowCORS then
    AResponse.SetCustomHeader('Access-Control-Allow-Origin','*');
  if not CheckListParams(aRequest,aResponse) then
    exit;
  if not InitSearch(aResponse) then
    exit;
  ConfigWordList(aRequest,aContaining,aPartial,aSimple);
  FSearch.GetAvailableWords(aList,aContaining,aPartial);
  J:=TJSONObject.Create;
  try
    if FIncludeMetadata then
      J.Add('metaData',FWordsMetadata.Clone);
    A:=TJSONArray.Create;
    if aSimple then
      For W in aList do
        A.Add(W)
    else
      begin
      For I:=0 to Length(aList)-1 do
        A.Add(TJSONObject.Create(['id',I+1,'word',aList[i]]));
      end;
    J.Add('data',A);
    SendJSON(J,aResponse);
  finally
    J.Free;
  end;
end;

initialization

end.

